@@ -1,21 +1,21 @@
 # -*- coding: utf-8 -*-
 from copy import copy
 
+import django
 from django.conf import settings
 from django.forms.formsets import BaseFormSet
 from django.template import Context
 from django.template.loader import get_template
-from django.utils.functional import memoize
 from django import template
 
 from crispy_forms.helper import FormHelper
+from crispy_forms.compatibility import lru_cache, string_types
 
 register = template.Library()
 # We import the filters, so they are available when doing load crispy_forms_tags
 from crispy_forms.templatetags.crispy_forms_filters import *
 
-TEMPLATE_PACK = getattr(settings, 'CRISPY_TEMPLATE_PACK', 'bootstrap')
-ALLOWED_TEMPLATE_PACKS = getattr(settings, 'CRISPY_ALLOWED_TEMPLATE_PACKS', ('bootstrap', 'uni_form'))
+from crispy_forms.utils import TEMPLATE_PACK, get_template_pack
 
 
 class ForLoopSimulator(object):
@@ -76,17 +76,17 @@ class BasicNode(template.Node):
     both the form object and parses out the helper string into attributes
     that templates can easily handle.
     """
-    def __init__(self, form, helper, template_pack=TEMPLATE_PACK):
+    def __init__(self, form, helper, template_pack=None):
         self.form = form
         if helper is not None:
             self.helper = helper
         else:
             self.helper = None
-        self.template_pack = template_pack
+        self.template_pack = template_pack or get_template_pack()
 
     def get_render(self, context):
         """
-        Returns a `Context` object with all the necesarry stuff for rendering the form
+        Returns a `Context` object with all the necessary stuff for rendering the form
 
         :param context: `django.template.Context` variable holding the context for the node
 
@@ -112,6 +112,13 @@ class BasicNode(template.Node):
             # This allows us to have simplified tag syntax: {% crispy form %}
             helper = FormHelper() if not hasattr(actual_form, 'helper') else actual_form.helper
 
+        # use template_pack from helper, if defined
+        try:
+            if helper.template_pack:
+                self.template_pack = helper.template_pack
+        except AttributeError:
+            pass
+
         self.actual_helper = helper
 
         # We get the response dictionary
@@ -143,7 +150,6 @@ class BasicNode(template.Node):
         """
         Returns a dictionary with all the parameters necessary to render the form/formset in a template.
 
-        :param attrs: Dictionary with the helper's attributes used for rendering the form/formset
         :param context: `django.template.Context` for the node
         :param is_formset: Boolean value. If set to True, indicates we are working with a formset.
         """
@@ -157,7 +163,7 @@ class BasicNode(template.Node):
 
         # We take form/formset parameters from attrs if they are set, otherwise we use defaults
         response_dict = {
-            'template_pack': settings.CRISPY_TEMPLATE_PACK,
+            'template_pack': self.template_pack,
             '%s_action' % form_type: attrs['attrs'].get("action", ''),
             '%s_method' % form_type: attrs.get("form_method", 'post'),
             '%s_tag' % form_type: attrs.get("form_tag", True),
@@ -179,6 +185,7 @@ class BasicNode(template.Node):
             'label_class': attrs.get('label_class', ''),
             'label_size': attrs.get('label_size', 0),
             'field_class': attrs.get('field_class', ''),
+            'include_media': attrs.get('include_media', True),
         }
 
         # Handles custom attributes added to helpers
@@ -192,14 +199,14 @@ class BasicNode(template.Node):
         return response_dict
 
 
+@lru_cache()
 def whole_uni_formset_template(template_pack=TEMPLATE_PACK):
     return get_template('%s/whole_uni_formset.html' % template_pack)
-whole_uni_formset_template = memoize(whole_uni_formset_template, {}, 1)
 
 
+@lru_cache()
 def whole_uni_form_template(template_pack=TEMPLATE_PACK):
     return get_template('%s/whole_uni_form.html' % template_pack)
-whole_uni_form_template = memoize(whole_uni_form_template, {}, 1)
 
 
 class CrispyFormNode(BasicNode):
@@ -214,6 +221,9 @@ class CrispyFormNode(BasicNode):
             else:
                 template = whole_uni_form_template(self.template_pack)
 
+        if django.VERSION >= (1, 8):
+            c = c.flatten()
+
         return template.render(c)
 
 
@@ -228,7 +238,7 @@ def do_uni_form(parser, token):
 
     Usage::
 
-        {% include crispy_tags %}
+        {% load crispy_tags %}
         {% crispy form form.helper %}
 
     You can also provide the template pack as the third argument::
@@ -243,32 +253,41 @@ def do_uni_form(parser, token):
     token = token.split_contents()
     form = token.pop(1)
 
+    helper = None
+    template_pack = "'%s'" % get_template_pack()
+
+    # {% crispy form helper %}
     try:
         helper = token.pop(1)
     except IndexError:
+        pass
+
+    # {% crispy form helper 'bootstrap' %}
+    try:
+        template_pack = token.pop(1)
+    except IndexError:
+        pass
+
+    # {% crispy form 'bootstrap' %}
+    if (
+        helper is not None and
+        isinstance(helper, string_types) and
+        ("'" in helper or '"' in helper)
+    ):
+        template_pack = helper
         helper = None
 
-    try:
-        # {% crispy form helper 'bootstrap' %}
-        template_pack = token.pop(1)
-        if not (template_pack[0] == template_pack[-1] and template_pack[0] in ('"', "'")):
-            raise template.TemplateSyntaxError("crispy tag's template_pack argument should be in quotes")
-
+    if template_pack is not None:
         template_pack = template_pack[1:-1]
+        ALLOWED_TEMPLATE_PACKS = getattr(
+            settings,
+            'CRISPY_ALLOWED_TEMPLATE_PACKS',
+            ('bootstrap', 'uni_form', 'bootstrap3', 'bootstrap4')
+        )
         if template_pack not in ALLOWED_TEMPLATE_PACKS:
-            raise template.TemplateSyntaxError("crispy tag's template_pack argument should be \
-                in " + str(ALLOWED_TEMPLATE_PACKS))
-    except IndexError:
-        # {% crispy form 'bootstrap' %}
-        # ('"'bootstrap'"', '"'uni_form'"','"'"bootstrap'"'", '"'"uni_form"'")
-        if (
-            helper in ['"%s"' % x for x in ALLOWED_TEMPLATE_PACKS] + \
-            ["'%s'" % x for x in ALLOWED_TEMPLATE_PACKS]
-        ):
-            template_pack = helper[1:-1]
-            helper = None
-        # {% crispy form helper %} OR {% crispy form %}
-        else:
-            template_pack = TEMPLATE_PACK
+            raise template.TemplateSyntaxError(
+                "crispy tag's template_pack argument should be in %s" %
+                str(ALLOWED_TEMPLATE_PACKS)
+            )
 
     return CrispyFormNode(form, helper, template_pack=template_pack)
